@@ -1,7 +1,11 @@
 #' @rdname prepending_uids
+#' @param outputPrefix,outputSuffix The prefix and suffix to add to the
+#' filenames when writing the processed files to disk.
 #' @export
 prepend_ids_to_sources <- function(input,
                                    output = NULL,
+                                   outputPrefix = "",
+                                   outputSuffix = "_withUIDs",
                                    origin=Sys.time(),
                                    preventOverwriting=rock::opts$get(preventOverwriting),
                                    encoding=rock::opts$get(encoding),
@@ -18,18 +22,20 @@ prepend_ids_to_sources <- function(input,
     stop("Only specify a single string as 'output'!");
   }
 
-  if (!dir.exists(input)) {
-    stop("Directory provided to read from ('",
-         input,
-         "') does not exist!");
-  }
-
-  if (!dir.exists(output)) {
-    warning("Directory provided to write to ('",
-            output,
-            "') does not exist - creating it!");
-    dir.create(output,
-               recursive = TRUE);
+  if (tolower(output) == "same") {
+    if ((is.null(outputPrefix) || (nchar(outputPrefix) == 0)) &&
+        (is.null(outputSuffix) || (nchar(outputSuffix) == 0))) {
+      stop("If writing the output to the same directory, you must specify ",
+           "an outputPrefix and/or an outputSuffix!");
+    }
+  } else {
+    if (!dir.exists(output)) {
+      warning("Directory provided to write to ('",
+              output,
+              "') does not exist - creating it!");
+      dir.create(output,
+                 recursive = TRUE);
+    }
   }
 
   rawSourceFiles <-
@@ -38,10 +44,25 @@ prepend_ids_to_sources <- function(input,
 
   res <- character();
   for (filename in rawSourceFiles) {
+    newFilename <-
+      paste0(outputPrefix,
+             sub("^(.*)\\.[a-zA-Z0-9]+$",
+                 "\\1",
+                 basename(filename)),
+             outputSuffix,
+             ".rock");
+    if (tolower(output) == "same") {
+      newFileDir <-
+        dirname(filename);
+    } else {
+      newFileDir <-
+        output;
+    }
+
     tmp <-
       prepend_ids_to_source(input = filename,
-                            output = file.path(output,
-                                               basename(filename)),
+                            output = file.path(newFileDir,
+                                               newFilename),
                             preventOverwriting = preventOverwriting,
                             origin=origin,
                             silent=silent);
